from typing import List
from bitarray import bitarray, frozenbitarray as fbarray
from bitarray.util import zeros as bazeros
from tqdm import tqdm


def topological_sorting(elements: List[fbarray]) -> (List[fbarray], List[int]):
    """Sort the list of `elements` by cardinality and lexicographical order

    In the output, the first element is the smallest one, the last one is the biggest

    Parameters
    ----------
    elements:
        List of bitarrays to sort (e.g. the list of intents)

    Returns
    -------
    elements_sorted: List[fbarray]
    orig_to_topsort_indices_map: List[int]
        THe mapping from the original indices to the indices of topologically sorted array.
    """
    ars_topsort = sorted(elements, key=lambda el: (el.count(), tuple(el.itersearch(True))))

    el_idx_map = {el: i for i, el in enumerate(ars_topsort)}
    orig_to_topsort_indices_map = [el_idx_map[el] for el in elements]
    return ars_topsort, orig_to_topsort_indices_map


def test_topologically_sorted(elements: List[fbarray]) -> bool:
    """Test if the list of `elements` is topologically sorted (from the smallest to the biggest element)

    One can obtain topologically sorted list of elements with ``topological_sorting(elements)`` function
    """
    return all(a.count() <= b.count() for a, b in zip(elements, elements[1:]))


def inverse_order(order: List[fbarray]) -> List[fbarray]:
    """Reverse the given order. So that new_order[i][j] = order[j][i]"""
    inversed = [bazeros(len(order[0])) for _ in order]
    for el_i, ordered in enumerate(order):
        for el_j in ordered.itersearch(True):
            inversed[el_j][el_i] = True

    return [fbarray(ordered) for ordered in inversed]


def sort_intents_inclusion(intents: List[fbarray], use_tqdm=False, return_transitive_order: bool = False)\
        -> List[fbarray] or (List[fbarray], List[fbarray]):
    """Returns indices of the smallest intents bigger than each intent. Optionally, return the transitive closure

    Returned is the list `lattice` so that intent with index `i` is included in intents with indices `lattice[i]`
    """
    assert test_topologically_sorted(intents), 'The `intents` list should be topologically sorted by ascending order'

    n_intents, n_attrs = len(intents), len(intents[0])
    zero_intents = fbarray(bazeros(n_intents))
    all_attrs = ~fbarray(bazeros(n_attrs))

    lattice = [fbarray(bazeros(n_intents))] * len(intents)
    trans_lattice = [fbarray(bazeros(n_intents))] * len(intents)

    attrs_descendants = [bitarray(zero_intents) for _ in range(n_attrs)]
    for intent_i, intent in enumerate(intents):
        for m in intent.itersearch(True):
            attrs_descendants[m][intent_i] = True

    for intent_i in tqdm(range(n_intents-1, -1, -1), disable=not use_tqdm, desc='Sorting intents'):
        intent = intents[intent_i]

        common_descendants = bitarray(~zero_intents)
        for m in intent.itersearch(True):
            common_descendants &= attrs_descendants[m]

        children = bitarray(zero_intents)
        for new_m in (all_attrs & ~intent).itersearch(True):
            meet_idx = (common_descendants & attrs_descendants[new_m]).find(True)
            children[meet_idx] = True

        trans_children = bitarray(zero_intents)
        for child in children.itersearch(True):
            trans_children |= trans_lattice[child]
        trans_lattice[intent_i] = fbarray(children | trans_children)
        lattice[intent_i] = fbarray(children & ~trans_children)

    if return_transitive_order:
        return lattice, trans_lattice
    return lattice


def close_transitive_subsumption(subsumption_list: List[fbarray]) -> List[fbarray]:
    """Transitive closure over subsumption operation

    So that in the returned list `trans_subsumption_list`,
    `trans_subsumption_list[i]` contains indices of all intents, smaller that intent with index `i`
    """
    assert all([max(subsumed.itersearch(True)) < i for i, subsumed in enumerate(subsumption_list) if subsumed.any()]), \
        "`subsumption_list` relation should be defined on a list, topologically sorted by descending order." \
        " So all subsumption_list of element `i` should have smaller indices"

    trans_subsumption_list = []
    for subsumed in subsumption_list:
        trans_subsumed = bitarray(subsumed)
        for el_i in subsumed.itersearch(True):
            trans_subsumed |= trans_subsumption_list[el_i]
        trans_subsumption_list.append(fbarray(trans_subsumed))
    return trans_subsumption_list


def open_transitive_subsumption(trans_subsumption_list: List[fbarray]) -> List[fbarray]:
    """Convert transitive closure of subsumption relation into the subsumption relation

    So that in the returned list `subsumption_list`,
    `subsumption_list[i]` contains indices of the biggest intents, smaller that intent with index `i`
    """
    assert all([
        max(subsumed.itersearch(True)) < i for i, subsumed in enumerate(trans_subsumption_list) if subsumed.any()
    ]),\
        "`trans_subsumption_list` relation should be defined on a list, topologically sorted by descending order." \
        " So all trans_subsumption_list of element `i` should have smaller indices"

    subsumption_list = []
    for trans_subsumed in trans_subsumption_list:
        subsumed = bitarray(trans_subsumed)
        for el_i in list(subsumed.itersearch(True))[::-1]:
            if subsumed[el_i]:
                subsumed &= ~trans_subsumption_list[el_i]
        subsumption_list.append(fbarray(subsumed))

    return subsumption_list


def drop_transitive_subsumption(subsumption_list: List[fbarray]) -> List[fbarray]:
    """Delete pairs with transitive relation from the given order

    If there are i,j,k such that all (subsumption_list[i][j], subsumption_list[j][k], subsumption_list[i,k]) are True,
    then set subsumption_list[i][k] = False.
    """
    return open_transitive_subsumption(close_transitive_subsumption(subsumption_list))



from typing import List, Dict, Iterator, Iterable

from .order import topological_sorting
from .base_functions import isets2bas, bas2isets

from skmine.itemsets import LCM
from bitarray import bitarray, frozenbitarray as fbarray
from bitarray.util import zeros as bazeros
from collections import deque


def list_intents_via_LCM(itemsets: List[fbarray], min_supp: float = 1, n_jobs: int = 1) -> List[fbarray]:
    """Get the list of intents by running LCM algorithm from scikit-mine

    Parameters
    ----------
    itemsets:
        The list of itemsets representing the dataset
    min_supp:
        Minimal support for the intent
    n_jobs:
        Number of jobs for LCM algorithm

    Returns
    -------
    intents:
        THe found intents

    """
    n_attrs = len(itemsets[0])

    lcm = LCM(min_supp=min_supp, n_jobs=n_jobs)
    intents = lcm.fit_transform(list(bas2isets(itemsets)))['itemset']
    intents = list(isets2bas(intents, n_attrs))
    intents = topological_sorting(intents)[0]

    biggest_intent = ~bazeros(n_attrs)
    if intents[-1] != biggest_intent:
        intents.append(fbarray(biggest_intent))

    smallest_intent = ~bazeros(n_attrs)
    for itset in itemsets:
        smallest_intent &= itset
        if not smallest_intent.any():
            break
    if intents[0] != smallest_intent:
        intents.insert(0, fbarray(smallest_intent))
    return intents


def list_attribute_concepts(intents: List[fbarray]) -> List[int]:
    """Get the indices of `intents` selected by each sole attribute"""
    assert (len(a) <= len(b) for a, b in zip(intents, intents[1:])),\
        'The list of `intents` should be topologically sorted. With the cardinality minimal intent being at the start'

    attr_concepts = [-1] * len(intents[0])
    found_attrs = bazeros(len(intents[0]))
    for intent_i, intent in enumerate(intents):
        for m in intent.itersearch(True):
            if attr_concepts[m] == -1:
                attr_concepts[m] = intent_i
        found_attrs |= intent
        if found_attrs.all():
            break

    return attr_concepts


def iter_equivalence_class(attribute_extents: List[fbarray], intent: fbarray = None) -> Iterator[fbarray]:
    """Iterate subsets of attributes from equivalence class

    The output equivalence class goes from the maximal subsets of attributes to the smallest ones.
    Equivalent subsets of attributes are the ones that describe the same subset of objects.


    Parameters
    ----------
    attribute_extents:
        The list of objects described by each specific attribute (converted to bitarrays)
    intent:
        Intent to compute equivalence class for. If None is passed, Iterate equivalence class of all attributes

    Returns
    -------
    Iterator[frozenbitarray]:
        Iterator over bitarrays representing equivalent subsets of attributes

    """
    N_OBJS, N_ATTRS = len(attribute_extents[0]), len(attribute_extents)

    intent = bazeros(N_ATTRS) if intent is None else intent

    def conjunct_extent(premise: fbarray) -> fbarray:
        res = ~bazeros(N_OBJS)
        for m in premise.itersearch(True):
            res &= attribute_extents[m]
            if not res.any():
                break

        return fbarray(res)

    total_extent = conjunct_extent(intent)
    stack = [[m] for m in intent.itersearch(True)][::-1]

    yield intent
    while stack:
        attrs_to_remove = stack.pop(0)
        last_attr = attrs_to_remove[-1]

        attrs_to_eval = bitarray(intent)
        for m in attrs_to_remove:
            attrs_to_eval[m] = False
        attrs_to_eval = fbarray(attrs_to_eval)

        conj = conjunct_extent(attrs_to_eval)
        if conj != total_extent:
            continue

        # conj == total_extent
        yield attrs_to_eval
        stack += [attrs_to_remove+[m] for m in intent.itersearch(True) if m > last_attr][::-1]


def list_keys_via_eqclass(equiv_class: Iterable[fbarray]) -> List[fbarray]:
    """List minimal subsets from given equivalence class"""
    potent_keys = []
    for new_key in equiv_class:
        potent_keys = [key for key in potent_keys if new_key & key != new_key]
        potent_keys.append(new_key)
    return potent_keys


def list_passkeys_via_eqclass(equiv_class: Iterable[fbarray]) -> List[fbarray]:
    """List subsets of minimal size from given equivalence class"""
    passkeys = []
    for descr in equiv_class:
        if not passkeys or descr.count() == passkeys[-1].count():
            passkeys.append(descr)

        if descr.count() < passkeys[-1].count():
            passkeys = [descr]
    return passkeys


def list_keys(intents: List[fbarray], only_passkeys: bool = False) -> Dict[fbarray, int]:
    """List all keys for all intents (i.e. minimal subsets of attributes selecting specific subsets of objects)"""
    assert all(a.count() <= b.count() for a, b in zip(intents, intents[1:])), \
        'The `intents` list should be topologically sorted by ascending order'

    n_attrs, n_intents = len(intents[-1]), len(intents)
    attrs_descendants = [bazeros(n_intents) for _ in range(n_attrs)]
    for intent_i, intent in enumerate(intents):
        for m in intent.itersearch(True):
            attrs_descendants[m][intent_i] = True

    # assuming that every subset of a key is a key => extending not-a-key cannot result in a key
    # and every subset of a passkey is a passkey

    single_attrs = list(isets2bas([[m] for m in range(n_attrs)], n_attrs))

    if only_passkeys:
        passkey_sizes = [n_attrs] * n_intents

    keys_dict = {fbarray(bazeros(n_attrs)): 0}
    attrs_to_test = deque([m_ba for m_ba in single_attrs])
    while attrs_to_test:
        attrs = attrs_to_test.popleft()
        attrs_indices = list(attrs.itersearch(True))

        if any(attrs & (~single_attrs[m]) not in keys_dict for m in attrs_indices):
            continue

        max_attr_idx = attrs_indices[-1] if attrs_indices else -1
        common_descendants = ~bazeros(n_intents)
        for m in attrs.itersearch(True):
            common_descendants &= attrs_descendants[m]

        meet_intent_idx = common_descendants.find(True)

        if only_passkeys:
            # `attrs` is not a passkey because of the size
            if passkey_sizes[meet_intent_idx] < attrs.count():
                continue

        # if subset of attrs is not a key, or a key of the same intent
        if any(keys_dict[attrs & (~single_attrs[m])] == meet_intent_idx for m in attrs_indices):
            continue

        keys_dict[attrs] = meet_intent_idx
        if only_passkeys:
            passkey_sizes[meet_intent_idx] = attrs.count()
        if meet_intent_idx != n_intents-1:
            attrs_to_test.extend([attrs | m_ba for m_ba in single_attrs[max_attr_idx+1:]])

    return keys_dict


def list_passkeys(intents: List[fbarray]) -> Dict[fbarray, int]:
    """List all passkeys for all intents

     (i.e. subsets of attributes of minimal size selecting specific subsets of objects)"""
    return list_keys(intents, only_passkeys=True)



import pytest

from caspailleur import order, base_functions as bfunc


def test_topological_sorting():
    elements = [frozenset(x) for x in [[], [1], [2], [0, 1], [0, 1, 2]]]
    elements_ba = list(bfunc.isets2bas(elements, 5))
    assert order.topological_sorting(elements_ba) == (elements_ba, [0, 1, 2, 3, 4])

    idxs_unordered = [1, 0, 3, 2, 4]
    elements_unordered = [elements_ba[i] for i in idxs_unordered]
    orig_to_topsort_map_true = [idxs_unordered.index(i) for i in range(len(idxs_unordered))]
    assert order.topological_sorting(elements_unordered) == (elements_ba, orig_to_topsort_map_true)


def test_inverse_order():
    inclusion_order = [set(), {0}, {1}, {0}, {2, 3}]
    subsumption_order_true = [{1, 3}, {2}, {4}, {4}, set()]

    inclusion_order_ba = list(bfunc.isets2bas(inclusion_order, 5))
    subsumption_order_true_ba = list(bfunc.isets2bas(subsumption_order_true, 5))

    assert order.inverse_order(inclusion_order_ba) == subsumption_order_true_ba
    assert order.inverse_order(order.inverse_order(inclusion_order_ba)) == inclusion_order_ba


def test_sort_intents_by_inclusion():
    elements = [set(), {1}, {2}, {0, 1}, {0, 1, 2}]
    elements_ba = list(bfunc.isets2bas(elements, 3))

    inclusion_order_true = [{1, 2}, {3}, {4}, {4}, set()]
    inclusion_order_true_ba = list(bfunc.isets2bas(inclusion_order_true, 5))

    inclusion_order = order.sort_intents_inclusion(elements_ba)
    assert inclusion_order == inclusion_order_true_ba

    with pytest.raises(AssertionError):
        order.sort_intents_inclusion(elements_ba[::-1])


def test_transitive_subsumption():
    subsumption_order_true = [set(), {0}, {0}, {1}, {2, 3}]
    transitive_order_true = [set(), {0}, {0}, {0, 1}, {0, 1, 2, 3}]
    semitransitive_order_true = [set(), {0}, {0}, {0, 1}, {0, 2, 3}]

    subsumption_order_true = list(bfunc.isets2bas(subsumption_order_true, 5))
    transitive_order_true = list(bfunc.isets2bas(transitive_order_true, 5))
    semitransitive_order_true = list(bfunc.isets2bas(semitransitive_order_true, 5))

    assert order.close_transitive_subsumption(subsumption_order_true) == transitive_order_true
    assert order.open_transitive_subsumption(transitive_order_true) == subsumption_order_true
    assert order.drop_transitive_subsumption(semitransitive_order_true) == subsumption_order_true



from bitarray import frozenbitarray as fbarray
import numpy as np
import os

from caspailleur import base_functions as bfunc


def test_powerset():
    iterable = (1, 2, 3)
    pset_true = [[], [1], [2], [3], [1, 2], [1, 3], [2, 3], [1, 2, 3]]
    pset_true = [frozenset(ar) for ar in pset_true]

    pset = bfunc.powerset(iterable)
    assert list(pset) == pset_true


def test_is_subset_of():
    assert bfunc.is_subset_of(set(), {1, 2, 3})
    assert bfunc.is_subset_of({2, 5, 3}, {2, 1, 3, 5})
    assert not bfunc.is_subset_of({2, 1, 3, 5}, {1})
    assert bfunc.is_subset_of({1, 2, 3}, {1, 2, 3})
    assert not bfunc.is_subset_of({0, 1, 3}, {0, 1, 2})


def test_is_psubset_of():
    assert bfunc.is_psubset_of(set(), {1, 2, 3})
    assert bfunc.is_psubset_of({2, 5, 3}, {2, 1, 3, 5})
    assert not bfunc.is_psubset_of({2, 1, 3, 5}, {1})
    assert not bfunc.is_psubset_of({1, 2, 3}, {1, 2, 3})


def test_closure():
    crosses_per_columns = [
        {0, 1,    3},  # col 0
        {0, 1, 2   },  # col 1
        {0,    2   },  # col 2
        {0, 1, 2, 3},  # col 3
        set(),         # col 4
    ]

    assert list(bfunc.closure([0], crosses_per_columns)) == [0, 3]
    assert list(bfunc.closure([0, 1], crosses_per_columns)) == [0, 1, 3]
    assert list(bfunc.closure([1, 2], crosses_per_columns)) == [1, 2, 3]
    assert list(bfunc.closure([], crosses_per_columns)) == [3]


def test_np_ba_conversions():
    X = np.array([[True, True, False], [False, False, True]])
    bars_true = [fbarray(x) for x in X.tolist()]
    bars = bfunc.np2bas(X)
    assert bars == bars_true

    X_reconstr = bfunc.bas2np(bars)
    assert (X == X_reconstr).all()

    attr_exts_true = [fbarray([True, False]), fbarray([True, False]), fbarray([False, True])]
    assert bfunc.np2bas(X.T) == attr_exts_true


def test_iset_ba_conversions():
    iset, l = {0, 1}, 5
    ba = fbarray([True, True, False, False, False])

    assert list(bfunc.isets2bas([iset], l)) == [ba]
    assert list(bfunc.bas2isets([ba])) == [iset]
    assert list(bfunc.bas2isets(bfunc.isets2bas([iset], l))) == [iset]


def test_conversion_pipeline():
    X = np.array([[True, True, False], [False, False, True]])
    itemsets_true = [{0, 1}, {2}]
    itemsets = list(bfunc.bas2isets(bfunc.np2bas(X)))
    assert itemsets == itemsets_true

    X_reconstr = bfunc.bas2np(bfunc.isets2bas(itemsets, 3))
    assert (X == X_reconstr).all()


def test_save_load_barrays():
    bitarrays = [fbarray([True, False]), fbarray([True, False]), fbarray([False, True])]
    with open('tst.bal', 'wb') as file:
        bfunc.save_balist(file, bitarrays)
    with open('tst.bal', 'rb') as file:
        bitarrays_load = list(bfunc.load_balist(file))

    os.remove('tst.bal')

    assert bitarrays_load == bitarrays



import pytest
import numpy as np

from caspailleur import mine_equivalence_classes as mec
from caspailleur import base_functions as bfunc

from bitarray import frozenbitarray as fbarray
from bitarray.util import zeros as bazeros


def test_list_intents_via_LCM():
    itemsets = [
        {0, 3},
        {0, 2},
        {1, 2},
        {1, 2, 3}
    ]

    intents_true = list(bfunc.isets2bas([set(), {0}, {2}, {3}, {0, 2}, {0, 3}, {1, 2}, {1, 2, 3}, {0, 1, 2, 3, 4}], 5))

    intents = mec.list_intents_via_LCM(list(bfunc.isets2bas(itemsets, 5)))
    assert intents == intents_true

    K = np.array([[True, False, False, False, False, True], [False, True, False, False, False, False]])
    itemsets = bfunc.np2bas(K)
    intents_true = list(bfunc.isets2bas([set(), {1}, {0, 5}, {0, 1, 2, 3, 4, 5}], 6))

    intents = mec.list_intents_via_LCM(itemsets)
    assert intents == intents_true


def test_list_attribute_concepts():
    intents = [set(), {0}, {2}, {3}, {0, 2}, {0, 3}, {1, 2}, {1, 2, 3},  {0, 1, 2, 3, 4}]
    attr_concepts_true = [1, 6, 2, 3, 8]

    attr_concepts = mec.list_attribute_concepts(list(bfunc.isets2bas(intents, 5)))
    assert attr_concepts == attr_concepts_true


def tests_iter_equivalence_class():
    K = np.array([
        [True, False, False, True, False],
        [True, False, True, False, False],
        [False, True, True, False, False],
        [False, True, True, True, False],
    ])
    attr_extents = list(bfunc.np2bas(K.T))
    eq_class_true = [
        {0, 1, 2, 3, 4}, {0, 1, 2, 3}, {0, 1, 2, 4}, {0, 1, 3, 4}, {0, 2, 3, 4}, {1, 2, 3, 4},
        {0, 1, 2}, {0, 1, 3}, {0, 1, 4}, {0, 2, 3}, {0, 2, 4}, {0, 3, 4}, {1, 2, 4}, {1, 3, 4}, {2, 3, 4},
        {0, 1}, {0, 4}, {1, 4}, {2, 4}, {3, 4}, {4}
    ]

    eq_class = list(mec.iter_equivalence_class(attr_extents, ~fbarray(bazeros(5))))
    assert eq_class == list(bfunc.isets2bas(eq_class_true, 5))


def test_list_keys_via_eqclass():
    eq_class = list(bfunc.isets2bas([
        {0, 1, 2, 3, 4}, {0, 1, 2, 3}, {0, 1, 2, 4}, {0, 1, 3, 4}, {0, 2, 3, 4}, {1, 2, 3, 4},
        {0, 1, 2}, {0, 1, 3}, {0, 1, 4}, {0, 2, 3}, {0, 2, 4}, {0, 3, 4}, {1, 2, 4}, {1, 3, 4}, {2, 3, 4},
        {0, 1}, {0, 4}, {1, 4}, {2, 4}, {3, 4}, {4}
    ], 5))
    keys_true = list(bfunc.isets2bas([{0, 2, 3}, {0, 1}, {4}], 5))

    keys = mec.list_keys_via_eqclass(eq_class)
    assert keys == keys_true


def test_list_passkeys_via_keys():
    keys = [{0, 2, 3}, {0, 1}, {4}, {0, 2, 4}]
    pkeys_true = [{4}]

    pkeys = mec.list_passkeys_via_eqclass(bfunc.isets2bas(keys, 5))
    assert pkeys == list(bfunc.isets2bas(pkeys_true, 5))


def test_list_keys():
    n_attrs = 5
    intents = [set(), {0}, {2}, {3}, {0, 2}, {0, 3}, {1, 2}, {1, 2, 3}, {0, 1, 2, 3, 4}]
    intents = list(bfunc.isets2bas(intents, n_attrs))

    keys_true = [
        (set(), 0),
        ({0}, 1), ({1}, 6), ({2}, 2), ({3}, 3), ({4}, 8),
        ({0, 1}, 8), ({0, 2}, 4), ({0, 3}, 5), ({1, 3}, 7), ({2, 3}, 7),
        ({0, 2, 3}, 8)
    ]
    keys_true = {next(bfunc.isets2bas([key], n_attrs)): intent_i for key, intent_i in keys_true}

    keys = mec.list_keys(intents)
    assert keys == keys_true

    with pytest.raises(AssertionError):
        mec.list_keys(intents[::-1])


def test_list_passkeys():
    n_attrs = 5
    intents = [set(), {0}, {2}, {3}, {0, 2}, {0, 3}, {1, 2}, {1, 2, 3}, {0, 1, 2, 3, 4}]
    intents = list(bfunc.isets2bas(intents, n_attrs))

    pkeys_true = [
        (set(), 0),
        ({0}, 1), ({1}, 6), ({2}, 2), ({3}, 3), ({4}, 8),
        ({0, 2}, 4), ({0, 3}, 5), ({1, 3}, 7), ({2, 3}, 7),
    ]
    pkeys_true = {next(bfunc.isets2bas([key], n_attrs)): intent_i for key, intent_i in pkeys_true}

    pkeys = mec.list_passkeys(intents)
    assert pkeys == pkeys_true

    with pytest.raises(AssertionError):
        mec.list_passkeys(intents[::-1])



from .base_functions import np2bas, bas2np
from .implication_bases import iter_proper_premises_via_keys, list_pseudo_intents_via_keys
from .mine_equivalence_classes import list_intents_via_LCM, list_keys, list_passkeys
from .order import sort_intents_inclusion, inverse_order
from .indices import linearity_index, distributivity_index

from .orchestrator import explore_data

import toml
__version__ = toml.load('pyproject.toml')['project']['version']






from caspailleur import implication_bases as impbas
from caspailleur import mine_equivalence_classes as mec
from caspailleur import base_functions as bfunc


def test_list_proper_premises_via_keys():
    n_attrs = 5
    intents = [set(), {0}, {2}, {3}, {0, 2}, {0, 3}, {1, 2}, {1, 2, 3}, {0, 1, 2, 3, 4}]
    intents = list(bfunc.isets2bas(intents, n_attrs))
    keys_to_intents = mec.list_keys(intents)

    pprems_true = list(bfunc.isets2bas([{1}, {4}, {2, 3}, {0, 1}, {0, 2, 3}], 5))

    pprems = dict(impbas.iter_proper_premises_via_keys(intents, keys_to_intents))
    assert set(pprems) == set(pprems_true)


def test_list_pseudo_intents_via_keys():
    pintents_true = [{1}, {0, 1, 2}, {2, 3}, {4}]
    intents = [set(), {0}, {2}, {3}, {0, 2}, {0, 3}, {1, 2}, {1, 2, 3}, {0, 1, 2, 3, 4}]
    keys_intent_map = [
        ([], 0), ([0], 1), ([1], 6), ([2], 2), ([3], 3), ([4], 8),
        ([0, 1], 8), ([0, 2], 4), ([0, 3], 5), ([1, 3], 7), ([2, 3], 7), ([0, 2, 3], 8)
    ]
    keys_intent_map = dict([(frozenset(k), v) for k, v in keys_intent_map])

    intents = list(bfunc.isets2bas(intents, 5))
    pintents_true = list(bfunc.isets2bas(pintents_true, 5))
    keys_intent_map = dict(zip(bfunc.isets2bas(keys_intent_map.keys(), 5), keys_intent_map.values()))

    pintents = impbas.list_pseudo_intents_via_keys(keys_intent_map.items(), intents)
    assert {pintent for pintent, _ in pintents} == set(pintents_true)



from typing import List, Dict, Tuple, Iterator, Iterable
from bitarray import frozenbitarray as fbarray, bitarray
from tqdm import tqdm
from caspailleur.order import test_topologically_sorted


def saturate(premise: fbarray, impls: List[Tuple[fbarray, int]], intents: List[fbarray]) -> fbarray:
    """Extend `premise` with implications from `impl` and intents `intents`

    Parameters
    ----------
    premise: frozenbitarray
        bitarray to saturate with implications (`impls`) and intents
    impls:
        List of implications to saturate with. Each implication is a pair of a bitarray and an intent index.
        Intent serves as the conclusion of each implication.
    intents: List[frozenbitarray]
        List of intents

    Returns
    -------
    saturated_premise: frozenbitarray
        `Premise` saturated with implications and intents


    Examples
    --------
    intents = [ bitarray('0000'), bitarray('0110'), bitarray('1111') ]
    impls = [ (bitarray('0100'), 1),  (bitarray('0001'), 2) ]

    Then premise '0101' contains the left part of the first implication, so it should be extended by intent #1: '0111'.
    New premise '0111' contains the left part of the second implication, so it should be extended by intent #2: '1111'.

    saturate( bitarray('0101'), impls, intents ) --> bitarray('0101') | intents[1] | intents[2] = bitarray('1111')
    """
    new_closure, old_closure = bitarray(premise), None
    new_unused_impl, old_unused_impl = list(impls), None

    while old_closure != new_closure:
        old_closure, old_unused_impl = bitarray(new_closure), new_unused_impl
        new_unused_impl = []

        for old_prem, old_closure_i in old_unused_impl:
            if old_prem & new_closure == old_prem:
                new_closure |= intents[old_closure_i]
            else:
                new_unused_impl.append((old_prem, old_closure_i))
    return fbarray(new_closure)


def test_if_proper_premise_via_keys(
        key: fbarray, intent_idx: int, intents: List[fbarray], keys_prevsize: Dict[fbarray, int]
) -> bool:
    """Test if `key` is a proper premise given dict of keys of smaller size

    Parameters
    ----------
    key: frozenbitarray
        A key (bitarray) to test
    intent_idx: int
        Index of the intent which is closure of `key`
    intents: List[frozenbitarray]
        List of bitarrays representing intents (i.e. closed subset of attributes)
    keys_prevsize: Dict[frozenbitarray, int]
        Dictionary containing all keys of smaller sizes and intents, corresponding to them.
        Passing dictionary of keys of size = size(key) - 1 is enough for algorithm to work.

    Returns
    -------
    flg: bool
        A flag, whether `key` is a proper premise

    Size of key -- size(key) -- is the number of True values in it (i.e. key.count() if key is a bitarray).
    """
    intent = intents[intent_idx]
    if key == intent:
        return False

    if key.count() == 0:
        return True

    cumulative_key = bitarray(key)
    for m in key.itersearch(True):
        prekey = bitarray(key)
        prekey[m] = False

        cumulative_key |= intents[keys_prevsize[fbarray(prekey)]]
        if cumulative_key == intent:
            return False
    return True


def iter_proper_premises_via_keys(intents: List[fbarray], keys_to_intents: Dict[fbarray, int])\
        -> Iterator[Tuple[fbarray, int]]:
    """Obtain the set of proper premises given intents, intents parents relation, and keys

    Parameters
    ----------
    intents: List[frozenbitarray]
        list of closed descriptions (in the form of bitarrays)
    keys_to_intents: Dict[frozenbitarray, int]
        the dictionary of keys in the context and the indices of the corresponding intents

    Returns
    -------
    proper_premises: Iterator[Tuple[frozenbitarray, int]]
        Iterator with found proper premises

    """
    return (
        (key, intent_idx) for key, intent_idx in keys_to_intents.items()
        if test_if_proper_premise_via_keys(key, intent_idx, intents, keys_to_intents)
    )


def list_pseudo_intents_via_keys(
        keys_intent_map: Iterable[Tuple[fbarray, int]], intents: List[fbarray],
        use_tqdm: bool = False, n_keys: int = None
) -> List[Tuple[fbarray, int]]:
    """List pseudo-intents (and indices of their intents) based on keys (and indices of their intents)

    Parameters
    ----------
    keys_intent_map: Iterable[Tuple[frozenbitarray, int]]
        The list (or generator) of keys and intents indices, corresponding to them
    intents: List[frozenbitarray]
        The list of intents (i.e. closed subsets of attributes)
    use_tqdm: bool
        A flag, whether to visualize the algorithm progress with tqdm progressbar
    n_keys: int
        Total number of keys (for better tqdm progressbar visualisation)

    Returns
    -------
    pseudo_intents_map: List[Tuple[frozenbitarray, int]]
        List of pseudo-intents and the indices of corresponding intents

    """
    assert test_topologically_sorted(intents), 'The `intents` list should be topologically sorted by ascending order'

    def add_pintent(
            new_key: fbarray, new_pintent: fbarray, new_intent_i: int,
            pintents_list: List[Tuple[fbarray, fbarray, int]]
    ) -> List[Tuple[fbarray, fbarray, int]]:
        if not pintents_list or pintents_list[-1][1].count() < new_pintent.count():
            pintents_list.append((new_key, new_pintent, new_intent_i))
            return pintents_list

        for new_idx in range(len(pintents_list)-1, -1, -1):
            if pintents_list[new_idx][1] == new_pintent:
                return pintents_list
            if pintents_list[new_idx-1][1].count() < new_pintent.count():
                break
        pintents_list.insert(new_idx, (new_key, new_pintent, new_intent_i))

        for idx in range(new_idx, len(pintents_list)):
            if len(pintents_list) <= idx:
                break

            new_pintents = []
            for (pi_key, pintent, pi_intent_idx) in pintents_list[idx:]:
                resaturated = saturate(pi_key, [pi_data[1:] for pi_data in pintents_list[:idx]], intents)
                if resaturated != intents[pi_intent_idx]:
                    new_pintents.append((pi_key, resaturated, pi_intent_idx))
            pintents_list[idx:] = sorted(new_pintents, key=lambda pi_data: pi_data[1].count())

        return pintents_list

    pseudo_intents: List[Tuple[fbarray, fbarray, int]] = []
    for key, intent_i in tqdm(keys_intent_map, total=n_keys, disable=not use_tqdm, desc="Iterate p.intent candidates"):
        key_saturated = saturate(key, [pi_data[1:] for pi_data in pseudo_intents], intents)
        if key_saturated == intents[intent_i]:
            continue

        pseudo_intents = add_pintent(key, key_saturated, intent_i, pseudo_intents)

    return [tuple(pi_data[1:]) for pi_data in pseudo_intents]



from itertools import chain, combinations
from typing import Iterable, Iterator, List, FrozenSet, BinaryIO
import numpy.typing as npt

import numpy as np
from bitarray import frozenbitarray as fbarray, bitarray
from bitarray.util import zeros as bazeros


################
# Math notions #
################
def powerset(iterable: Iterable[int]) -> Iterable[FrozenSet[int]]:
    """Iterate through all subsets of elements of set `iterable`

    Examples
    --------
    powerset({1,2,3}) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)
    """
    s = list(iterable)
    return map(frozenset, chain.from_iterable(combinations(s, r) for r in range(len(s)+1)))


def is_subset_of(A: FrozenSet[int] or fbarray, B: FrozenSet[int] or fbarray) -> bool:
    """Test whether `A` is a subset of `B`"""
    return A & B == A


def is_psubset_of(A: FrozenSet[int] or fbarray, B: FrozenSet[int] or fbarray) -> bool:
    """Test whether `A` is a proper subset of `B`"""
    return (A & B == A) and A != B


def closure(D: Iterable[int], crosses_per_columns: List[FrozenSet[int]]) -> Iterator[int]:
    """Iterate indices of all columns whose intersection equals to the intersection of columns `D`

    Parameters
    ----------
    D: Iterable[int]
        Indices of some columns from `crosses_per_columns`
    crosses_per_columns: List[FrozenSet[int]]
        List of indices of 'True' rows for each column

    Returns
    -------
    intent: Iterator[int]
        Indices of All columns with the same intersection as `D`

    """
    n_rows = max(max(col) + 1 for col in crosses_per_columns if col)

    extent = set(range(n_rows))
    for m in D:
        extent &= crosses_per_columns[m]

    intent = (m for m, col in enumerate(crosses_per_columns) if is_subset_of(extent, col))
    return intent


##########################
# Basic type conversions #
##########################
def np2bas(X: npt.ArrayLike) -> List[fbarray]:
    """Convert numpy boolean matrix to the list of bitarrays (one per row of the matrix)"""
    return [fbarray(x) for x in X.tolist()]


def bas2np(barrays: Iterable[fbarray]) -> npt.ArrayLike:
    """Convert the list of bitarrays to numpy boolean matrix (bitarrays become rows in of the matrix)"""
    return np.vstack([ba.tolist() for ba in barrays]).astype(np.bool_)


def isets2bas(itemsets: Iterable[Iterable[int]], length: int) -> Iterator[fbarray]:
    """Convert the list of lists of indices of 'True' elements to bitarrays of given length

    Examples
    --------
    isets2bas([ [0, 1], [1,3,4] ], 5) --> [bitarray('01000'), bitarray('01011')]
    """
    for iset in itemsets:
        bar = bazeros(length)
        for m in iset:
            bar[m] = True
        yield fbarray(bar)


def bas2isets(bitarrays: Iterable[fbarray]) -> Iterator[FrozenSet[int]]:
    """Convert the list of bitarrays to the list of lists of indices of 'True' elements of the bitarrays

    Examples
    --------
    bas2isets([ bitarray('01000'), bitarray('01011') ]) --> [ [0,1], [1,3,4] ]
    """
    for bar in bitarrays:
        yield frozenset(bar.itersearch(True))


###########################
# Save and load functions #
###########################

def load_balist(file: BinaryIO) -> Iterator[bitarray]:
    """Load the list of bitarrays from binary file"""
    basize = b''
    while True:
        data = file.read(1)
        if data == b'n':
            break
        basize += data
    basize = int(basize.decode(encoding='utf-8'))
    basize_bytes = len(bazeros(basize).tobytes())

    while True:
        data = file.read(basize_bytes)
        if data == b'':
            break

        ba = bitarray()
        ba.frombytes(data)
        yield ba[:basize]


def save_balist(file: BinaryIO, bitarrays: List[bitarray]):
    """Save the list of bitarrays to the binary file (proposed file extension '.bal')"""
    basize = len(bitarrays[0])
    assert all(len(ba) == basize for ba in bitarrays), "All bitarrays should be of the same size"

    file.write(str(basize).encode(encoding='utf-8'))
    file.write(b'n')

    for ba in bitarrays:
        file.write(ba.tobytes())
